/*******************************************************************************
 * Copyright (c) 2004, 2006 IBM Corporation. 
 * All rights reserved. 
 *******************************************************************************/
package com.ibm.gallery.common.wizards;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.osgi.framework.Bundle;

import org.eclipse.ui.dialogs.IOverwriteQuery;

import com.ibm.ccl.ua.wizards.*;



public class ProjectImportOperation implements IRunnableWithProgress {
	
	// RATLC00530885 - There are problems on RedHat Linux when Java source files are
	// imported using the Eclipse ZipFileStructureProvider and ImportOperation, the
	// input stream created for reading the source files uses the default character
	// encoding. On Linux, this default encoding does not appear to be able to handle
	// the Java source files which uses the ISO-8859-1 character encoding. 
	//
	// To solve the problem, code was added to the method importFilesFromZip()
	// to use the correct encoding when importing java source files. The other source
	// file (plugin.xml, .project, *.emx) do not appear to have problems with the 
	// default encoding. After fixing the encoding problem, another problem was noticed. 
	// The previous logic would create the project and then import the ZIP file into that
	// project. If the wizard definition (in plugin.xml) did not define any natures
	// or references, the project is created without any. However, the ZIP file may
	// import a .project file with the natures/references defined. As a result, after 
	// the import, the project does not appear to have the natures but closing and 
	// reopening the project restores the natures from the .project file. The solution
	// is to unzip the files first and then create the project. If the wizard definition
	// does not define any natures or references, don't set them in the project descriptor.
	
	private IResource elementToOpen;
	private ProjectImportWizardPage[] fPages;
	// used for overwrite warning
	private IOverwriteQuery overwriteQuery;
	/**
	 * Constructor for ProjectImportOperation 
	 */ 
	public ProjectImportOperation(ProjectImportWizardPage[] pages, IOverwriteQuery newOverwriteQuery) {
		elementToOpen = null;
		fPages = pages;
		overwriteQuery = newOverwriteQuery;
	}
	
	/**
	 * Constructor for ProjectImportOperation
	 */
	public ProjectImportOperation(ProjectImportWizardPage[] pages) {
		elementToOpen = null;
		fPages = pages;
	}
	
	/*
	 * @see IRunnableWithProgress#run(IProgressMonitor)
	 */
	public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
		if (monitor == null) {
			monitor= new NullProgressMonitor();
		}
		try {
			monitor.beginTask(ProjectImportMessages.getString("ProjectImportOperation.op_desc"), fPages.length); //$NON-NLS-1$
			IWorkspaceRoot root= Activator.getWorkspace().getRoot();
			
			for (int i= 0; i < fPages.length; i++) {
				createProject(root, fPages[i], new SubProgressMonitor(monitor, 1));
			}
		} finally {
			monitor.done();
		}
	}		
	
	public IResource getElementToOpen() {
		return elementToOpen;
	}
	
	private void createProject(IWorkspaceRoot root, ProjectImportWizardPage page, IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
		IConfigurationElement desc = page.getConfigurationElement();
		IConfigurationElement[] imports = desc.getChildren("import"); //$NON-NLS-1$
		IConfigurationElement[] natures = desc.getChildren("nature"); //$NON-NLS-1$
		IConfigurationElement[] references= desc.getChildren("references"); //$NON-NLS-1$
		int nImports = (imports == null) ? 0 : imports.length;
		int nNatures = (natures == null) ? 0 : natures.length;
		int nReferences= (references == null) ? 0 : references.length;
		
		monitor.beginTask(ProjectImportMessages.getString("ProjectImportOperation.op_desc_proj"), nImports + 1); //$NON-NLS-1$

		String name= page.getName();
		String[] natureIds= new String[nNatures];
		for (int i= 0; i < nNatures; i++) {
			natureIds[i]= natures[i].getAttribute("id"); //$NON-NLS-1$
		}
		IProject[] referencedProjects= new IProject[nReferences];
		for (int i= 0; i < nReferences; i++) {
			referencedProjects[i]= root.getProject(references[i].getAttribute("id")); //$NON-NLS-1$
		}	

		// RATLC00530885 - Create a directory where the project files will be imported.
		File projectFolderFile = new File(Platform.getLocation().append(name).toOSString());
		projectFolderFile.mkdirs();
		
		// RATLC00530885 - Imports must be done before creating the project in case the
		// Zip file contains a .project file
		for (int i= 0; i < nImports; i++) {
			doImports(projectFolderFile, imports[i], new SubProgressMonitor(monitor, 1));
		}
		
		IProject proj= configNewProject(root, name, natureIds, referencedProjects, monitor);
		
		String open= desc.getAttribute("open"); //$NON-NLS-1$
		if (open != null && open.length() > 0) {
			IResource fileToOpen= proj.findMember(new Path(open));
			if (fileToOpen != null) {
				elementToOpen= fileToOpen;
			}
		}		
		
	}
	
	private IProject configNewProject(IWorkspaceRoot root, String name, String[] natureIds, IProject[] referencedProjects, IProgressMonitor monitor) throws InvocationTargetException {
		try {
			IProject project= root.getProject(name);
			if (!project.exists()) {
				project.create(null);
			}
			if (!project.isOpen()) {
				project.open(null);
			}
			IProjectDescription desc= project.getDescription();
			
			// RATLC00530885 - Only set nature IDs and references if supplied in wizard
			// definition extension, otherwise they will override an imported .project file.
			// Since the project already resides in the workspace, do not reset the
			// location in the project descriptor.

			boolean setDescriptor = false;
			if (natureIds.length > 0) {
				desc.setNatureIds(natureIds);
				setDescriptor = true;
			}
			if (referencedProjects.length > 0) {
				desc.setReferencedProjects(referencedProjects);
				setDescriptor = true;
			}
			
			if (setDescriptor) {
				project.setDescription(desc, new SubProgressMonitor(monitor, 1));
			}

			return project;
		} catch (CoreException e) {
			Activator.logError("ProjectImportOperation configNewProject() " , e);//$NON-NLS-2$//$NON-NLS-1$
			e.printStackTrace();
			throw new InvocationTargetException(e);
		}
	}
	
	private void doImports(File project, IConfigurationElement curr, IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
		try {
			File destDirectory;
			String name= curr.getAttribute("dest"); //$NON-NLS-1$
			if (name == null || name.length() == 0) {
				destDirectory= project;
			} else {
				IPath destPath = new Path(project.getAbsolutePath()).append(name);
				destDirectory = new File(destPath.toOSString());
				if (!destDirectory.exists()) {
					destDirectory.mkdirs();
				}
			}
			String importPath= curr.getAttribute("src"); //$NON-NLS-1$
			if (importPath == null) {
				importPath= ""; //$NON-NLS-1$
				Activator.logWarning("projectsetup descriptor: import missing"); //$NON-NLS-1$
				return;
			}
		
			ZipFile zipFile= getZipFileFromPluginDir(importPath, getContributingPlugin(curr));
			// RATLC00530885 - changed import call with number zip entries and the destination directory
			importFilesFromZip(zipFile, destDirectory, new SubProgressMonitor(monitor, zipFile.size()));
		// RATLC00530855 - changed CoreException in catch to just Exception	
		} catch (Exception e) {
			Activator.logWarning("ProjectImportOperation" + " doImports() " + e);//$NON-NLS-2$//$NON-NLS-1$
			e.printStackTrace();
			throw new InvocationTargetException(e);
		}
	}
	
	private String getContributingPlugin(IConfigurationElement configurationElement) {
		Object parent= configurationElement;
		while(parent != null) {
			if (parent instanceof IExtension)
				return ((IExtension)parent).getNamespaceIdentifier();
			parent= ((IConfigurationElement)parent).getParent();
		}
		return null;
	}

	private ZipFile getZipFileFromPluginDir(String pluginRelativePath, String symbolicName) throws CoreException {
		try {
			Bundle bundle = Platform.getBundle(symbolicName);
			URL starterURL = new URL(bundle.getEntry("/"), pluginRelativePath); //$NON-NLS-1$
			
			// deprecated return new ZipFile(Platform.asLocalURL(starterURL).getFile());
			return new ZipFile(FileLocator.toFileURL(starterURL).getFile());
			
		} catch (IOException e) {
			String message= pluginRelativePath + ": " + e.getMessage(); //$NON-NLS-1$
			Status status= new Status(IStatus.ERROR, Activator.PLUGIN_ID, IStatus.ERROR, message, e);
			throw new CoreException(status);
		}
	}
	
	private void importFilesFromZip(ZipFile srcZipFile, File destDirectory, IProgressMonitor monitor) throws IOException, FileNotFoundException, InterruptedException {		

		// The destination directory is passed as a File (no longer an IPath).
		/*
		 * Walk each element and extract it (ignore directories)
		 */
		try {
			Enumeration e = srcZipFile.entries();

			while (e.hasMoreElements()) {
				ZipEntry zipEntry = (ZipEntry) e.nextElement();
				File file = new File(destDirectory, zipEntry.getName());

				if (false == zipEntry.isDirectory()) {

					/*
					 * Copy files (and make sure parent directory exist)
					 */
					File parentFile = file.getParentFile();
					if (null != parentFile && false == parentFile.exists()) {
						parentFile.mkdirs();
					}

					Path path = new Path(file.getPath());
					if (path.getFileExtension() != null && path.getFileExtension().equals("java")) { //$NON-NLS-1$
						InputStreamReader is = null;
						OutputStreamWriter os = null;

						try {
							is = new InputStreamReader(srcZipFile
								.getInputStream(zipEntry), "ISO-8859-1"); //$NON-NLS-1$
							os = new OutputStreamWriter(new FileOutputStream(
								file), ResourcesPlugin.getEncoding());
							char[] buffer = new char[102400];
							while (true) {
								int len = is.read(buffer);
								if (len < 0)
									break;
								os.write(buffer, 0, len);
							}
						} finally {
							if (null != is) {
								is.close();
							}
							if (null != os) {
								os.close();
							}
						}
					} else {
						InputStream is = null;
						OutputStream os = null;

						try {
							is = srcZipFile.getInputStream(zipEntry);
							os = new FileOutputStream(file);
							
							byte[] buffer = new byte[102400];
							while (true) {
								int len = is.read(buffer);
								if (len < 0)
									break;
								os.write(buffer, 0, len);
							}
						} catch (Exception exp){
							//System.out.println("Exception in file import");
						}
						 
						 finally {
							if (null != is) {
								is.close();
							}
							if (null != os) {
								os.close();
							}
						}
					}
				}

				monitor.worked(1);

				if (monitor.isCanceled()) {
					throw new InterruptedException();
				}
			}
		} finally {
			srcZipFile.close();
			monitor.done();
		}
	}
}
